/*******************************************************************************
* SiRF_KillStaicNav:   NMEA 0183 & SiRF Binary Protocol library for Arduino    *
*                      example application: Disables the "Static Navigation"   *
*                      feature on Parallax PMB-648 GPS SiRF GPS Receivers.     * 
* Jacob Dilles <jdilles@gmu.edu>   November  2012     Rev2                     *
********************************************************************************/  

/* Import the SiRF Library */
#include <sirf.h>

/* 
    Declare our GPS Object.
    
    For this example, we assume there is a Parallax PMB-648 connected to
    an Arduino Mega board, as follows:
        Red     ->  +5V
        Black   ->  GND
        Yellow  ->  RX1
        Blue    ->  TX1
*/
SiRF gps;

void setup(){
    /* Hardware Serial for Debugging */
    Serial.begin(9600);
    
    /* Configure the SiRF object on HardwareSerial1, 4800 baud */
    gps.begin(Serial1, 4800);
   
    /* Give ArduinoIDE a chance to reprogram before spamming Serial */
    Serial.println("Starting...");
    delay(5000);
    
    /* 
      Wait for data, to see what mode the reciever is in 
    */
    Serial.println("Waiting for data...");
    Protocol mode;
    while( !(mode=gps.receive()) ) Serial.print('.');
    
    /* 
        Drop to binary mode. Ref: SiRF GPS Protocol Reference Manual Rev 1.30
            -> See "1.2.3 SetSerialPort" on pg 9 
    */
    Serial.print(("\nEntering Binary Mode..."));
    if(mode == NMEA) /* only need to do this if we're not in BINARY mode */
    if(gps.sendNMEA("$PSRF100,0,4800,8,1,0*")){
        Serial.println("Error sending NMEA command");
        return;
    }
    while(gps.receive() != BINARY) Serial.print('.');
    
    
    
    /* 
        Turn off static nav. Ref: SiRF Binary Protocol Reference Manual Rev2.4
            -> See "Static Navigation – Message ID 143" pg 2-16 
    */
    Serial.print("\nTurning off Static Nav (143d)...");
    if(gps.sendHEX("8F00")){
        Serial.println("Error sending binary command 143"); 
        return;
    }
    while(gps.receive() != BINARY) Serial.print('.');
    
    /* Turn off degraded mode - i.e. force a 3D fix */
    
    Serial.print("\nTurning off degrated mode...");
    if(gps.sendHEX("88"  /* message ID */
                   "0000" /* reserved */
                   "04"   /* dsiable degraded mode (see table 2-20) */
                   "0000" /* reserved */
                   "0000" /* User altitude - ignore */
                   "02"   /* Alt hold mode: Ignore user altitude - force 3d fix */
                   "00"   /* alt hold source - last computed */
                   "00"   /* reserved */
                   "00"   /* Degraded timeout = disable */
                   "00"   /* Dead-recokining timeout = disable */
                   "00"   /* track smoothing = disable */
                   )){
      Serial.println("Error sending binary command 136");                  
    }
    while(gps.receive() != BINARY) Serial.print('.');
    
    
    
    /* 
        Return to NMEA mode. 
          -> See "Switch To NMEA Protocol – Message ID 129" pg 2-6
    */
    Serial.print(("\nReturning to NMEA mode..."));
    if(gps.sendHEX(("8102" /* Switch To NMEA Protocol – Message ID 129 */
                   "0101" /* GGA - Global Positioning System Fix Data */
                   "0000" /* GLL - Geographic Position, Latitude / Longitude and time. */
                   "0101" /* GSA - GPS DOP and active satellites */
                   "0501" /* GSV - GPS Satellites in view */
                   "0101" /* RMC - Recommended minimum specific GPS/Transit data */
                   "0000" /* VTG - Track Made Good and Ground Speed */
                   "0001" /* MMS - ? */
                   "0001" /* EPE - ? */
                   "0001" /* ZDA - UTC, day, month, year, and local time zone. */
                   "0000" /* unused */
                   "12C0" /* 4800 baud */
                ))){
        Serial.println("Error sending binary command 129"); 
        return;
    }
    while(gps.receive() != NMEA) Serial.print('.');
    Serial.println("\nDone! Static Nav should be off");
    
}

Position waypoint[] = {
    Position( DMSH(38, 49, 40.15, N), DMSH(77, 18, 18.95, W) ),
    Position( DMSH(38, 49, 39.68, N), DMSH(77, 18, 20.09, W) ),
    Position( DMSH(38, 49, 39.31, N), DMSH(77, 18, 21.09, W) )
};

/** Just print out the GPRMC sentence and demonstare parsing of params */
void loop(){  
  GPRMC d;
  GPGSV s;
  Protocol mode = gps.receive();
  
  if(mode  == NMEA){
    
    //Serial.println(gps.sentence());
    if(gps.parseGPRMC(d)){
      Serial.print("GPRMC[");
      Serial.print(d.status);
      Serial.print("],");
      
      Position p = d.location();
// Also, you can get
//  Vector    d.velocity()
//  char      d.utc_year
//  char      d.utc_month
//  char      d.utc_day
//  char      d.utc_hour
//  char      d.utc_min
//  char      d.utc_sec

      Serial.print(p.lat, 6);
      Serial.print(",");
      Serial.print(p.lon, 6);

      if(d.status == 'A'){
         Vector wpv = p.vectorTo(waypoint[0]);
        
         Serial.print((",VEC,"));
         Serial.print(wpv.mag, 6); 
         Serial.print((","));
         Serial.print(wpv.dir, 6);
      }
      Serial.println();
      
       
    } else if ( gps.parseGPGSV(s) ){
        if(s.msg_num == 1)Serial.print("SNRs: ");      
        Serial.print(s.satellite[0].snr); Serial.print(" ");
        Serial.print(s.satellite[1].snr); Serial.print(" ");
        Serial.print(s.satellite[2].snr); Serial.print(" ");
        Serial.print(s.satellite[3].snr); Serial.print(" ");
        if(s.msg_num == s.num_msg)Serial.println();
    }
  }
  
}